@@ -1,6 +1,7 @@
 """ Filesystem wrapper that provides versioning capabilities through
     rdiff-backup.
 """
+from collections import namedtuple
 import hashlib
 import os
 import random
@@ -13,6 +14,7 @@ from fs.base import synchronize
 from fs.filelike import FileWrapper
 from fs.errors import OperationFailedError, ResourceNotFoundError
 from fs.path import relpath
+from fs.tempfs import TempFS
 
 from versioning_fs.errors import SnapshotError, VersionError
 from versioning_fs.hidefs import HideFS
@@ -20,6 +22,8 @@ from versioning_fs.hidefs import HideFS
 
 hasher = hashlib.sha256  # hashing function to use with backup paths
 
+VersionInfo = namedtuple('VersionInfo', ['timestamp',  'size'])
+
 
 def hash_path(path):
     """Returns a hash of a given path."""
@@ -79,6 +83,33 @@ class VersionInfoMixIn(object):
         info = {k+1: formatted_time(int(v)) for k, v in enumerate(versions)}
         return info
 
+    def list_sizes(self, path):
+        """Returns a dictionary containing sizes for each version of a path.
+        """
+        snap_dir = self.snapshot_snap_path(path)
+        command = ['rdiff-backup', '--parsable-output',
+                   '--list-increment-sizes', snap_dir]
+        process = Popen(command, stdout=PIPE, stderr=PIPE)
+        stdout = process.communicate()[0]
+
+        listing_file = StringIO(stdout)
+        if len(listing_file.readlines()) < 3:
+            return {}
+
+        listing_file.seek(0)
+
+        # skip the first two lines of output
+        for _ in range(2):
+            next(listing_file)
+
+        # generate a dictionary
+        sizes = dict()
+        for version, line in enumerate(reversed(listing_file.readlines())):
+            size = "%s %s" % (line.split()[5], line.split()[6])
+            sizes[version+1] = size
+
+        return sizes
+
 
 class VersioningFS(VersionInfoMixIn, HideFS):
     """ Versioning filesystem.
@@ -183,9 +214,7 @@ class VersioningFS(VersionInfoMixIn, HideFS):
                 process = Popen(command, stdout=PIPE, stderr=PIPE)
                 process.communicate()
 
-                dest_hash = hash_path(path)
-
-                file_path = os.path.join(temp_name, dest_hash)
+                file_path = os.path.join(temp_name, 'datafile')
                 open_file = self.tmp.open(file_path, mode=mode)
                 return VersionedFile(fs=self, file_object=open_file,
                                      mode=mode, temp_file=True,
@@ -254,31 +283,24 @@ class VersioningFS(VersionInfoMixIn, HideFS):
     def snapshot(self, path):
         """Takes a snapshot of an individual file."""
 
-        # relative to the mounted fs, what should be snapshotted and where
-        # should it go
-        snap_source_dir = self.snapshot_source(path)
-        snap_dest_dir = self.snapshot_snap_path(path)
+        # try grabbing the temp filesystem system path
+        temp_dir = None
+        if 'getsyspath' in dir(self.tmp):
+            temp_dir = self.tmp.getsyspath('/')
 
-        # create the directory where the snapshot will be taken from
-        if os.path.exists(snap_source_dir):
-            shutil.rmtree(snap_source_dir)
-        os.makedirs(snap_source_dir)
-        if not self.has_snapshot(path):
-            os.makedirs(snap_dest_dir)
+        # Create a temp file system to be snapshotted
+        temp_snapshot_fs = TempFS(temp_dir=temp_dir)
+        src_path = temp_snapshot_fs.getsyspath('/')
 
-        link_src = self.fs.getsyspath(path)
+        with self.fs.open(path, 'rb') as source_file:
+            with temp_snapshot_fs.open('datafile', 'wb') as temp_file:
+                shutil.copyfileobj(source_file, temp_file)
 
-        dest_hash = hash_path(path)
-        link_dst = os.path.join(snap_source_dir, dest_hash)
-
-        # hardlink the user file to a file inside a temp dir
-        os.link(link_src, link_dst)
-
-        src_path = os.path.join(self.__tmp.getsyspath('/'), snap_source_dir)
-        dest_path = snap_dest_dir
+         # snapshot destination directory
+        dest_dir = self.snapshot_snap_path(path)
 
         command = ['rdiff-backup', '--parsable-output', '--no-eas',
-                   '--no-file-statistics', '--no-acls', src_path, dest_path]
+                   '--no-file-statistics', '--no-acls', src_path, dest_dir]
 
         # speed up the tests
         if self.__testing:
@@ -296,8 +318,8 @@ class VersioningFS(VersionInfoMixIn, HideFS):
                 if not rule(stderr):
                     raise SnapshotError(stderr)
 
-        # remove  the intermediate directory
-        shutil.rmtree(snap_source_dir)
+        # close the temp snapshot filesystem
+        temp_snapshot_fs.close()
 
     def remove_versions_before(self, path, version):
         """Removes snapshots before a specified version.
@@ -363,15 +385,6 @@ class VersioningFS(VersionInfoMixIn, HideFS):
         save_snap_dir = os.path.join(backup_dir, dest_hash)
         return save_snap_dir
 
-    def snapshot_source(self, path):
-        """Returns the dir of the file to be snapshotted. This dir should
-           contain a hardlink to the original file in the user files
-           directory.
-        """
-
-        snap_dir = "%s.backup" % (self.snapshot_info_path(path))
-        return snap_dir
-
 
 class VersionedFile(FileWrapper):
     """File wrapper that notifies the versioning filesystem to take a
